package org.ezunit.evaluation;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.ezunit.evaluation.MatrixDataReader.Result;
import org.ezunit.evaluation.MatrixDataReader.Test;
import org.ezunit.evaluation.MatrixDataReader.UUT;

public class BDM {

	// methods
	public static Set<Set<Test>> getBlocks(Result input) {
		// if either no rows or no tests are given, do nothing
		if ( input.uuts.size() == 0 || input.tests.size() == 0 )
			return new HashSet<Set<Test>>();
	
		boolean[][] matrix = new boolean[input.uuts.size()][input.tests.size()];
		
		int rowIndex = 0, colIndex = 0;
		
		for ( UUT uut : input.uuts ) {
			for ( Test test : input.tests ) {
				matrix[rowIndex][colIndex] = uut.callingTests.contains( test );
				
				++colIndex;
			}

			++rowIndex;
			colIndex = 0;
		}
		
		// collect partitions
		List<Integer>[] sparseMatrix = getCoverageMatrix( matrix );
		List<List<Integer>> partitionsColIndices = extractPartitions( sparseMatrix );

		// create the result
		Set<Set<Test>> result = new HashSet<Set<Test>>();
		
		for ( List<Integer> cols : partitionsColIndices ) {
			HashSet<Test> tests = new HashSet<Test>();
			
			for ( Integer col : cols )
				tests.add( input.tests.get( col ) );
			
			if ( tests.size() > 0 )
				result.add( tests );
		}
		
		return result;
	}

	// private methods

	/**
	 * Returns a list of partition information, where each partition information is a
	 * list of the column indices, denoting the columns that form a partition.
	 * 
	 * @param sparseCoverageMatrix sparse coverage matrix to process
	 * @return partition list
	 */
	private static List<List<Integer>> extractPartitions( List<Integer>[] sparseCoverageMatrix ) {
		// find all intersections by successively creating the partitions and joining
		// new partitions with existing ones in case of an overlap;
		//
		// example:
		//
		// coverageMatrix[][] =
		//      t1 t2 t3
		//  m1  1 
		//  m2  1  1     
		//  m3     1  1 
		//  m4        1
		//
		// if the working order is [t1, t3, t2] (as it might be possible due to the unordered
		// nature of sets), there would be two clusters found in the first two steps: {t1} and {t3};
		// the next step, working on t2, requires the unification of {t1} and {t3} to the cluster {t1, t2, t3}
		List<List<Integer>> partitionIndices = new LinkedList<List<Integer>>();
		
		// for all test cases
		for ( int testIndex = 0; testIndex < sparseCoverageMatrix.length; ++testIndex ) {
			List<Integer> newPartition = new LinkedList<Integer>();
			newPartition.add( testIndex );

			// for all partitions
			loop: for ( Iterator<List<Integer>> partIt = partitionIndices.iterator(); partIt.hasNext(); ) {
				List<Integer> partition = partIt.next();
				
				List<Integer> testMUTs = sparseCoverageMatrix[testIndex];
				
				for ( Integer partitionTestIndex : partition )
					for ( Integer mut : sparseCoverageMatrix[partitionTestIndex] )
						if ( testMUTs.contains( mut ) ) {
							// add everything at the beginning so that the test methods will not be reversed
							newPartition.addAll( 0, partition ); 
							partIt.remove();
							continue loop;
						}
			}
			
			partitionIndices.add( newPartition );
		}
		
		return partitionIndices;
	}
	
	/**
	 * Returns the sparse coverage matrix according to the given informations.
	 * The result is stored in column form, i.e. the resulting array contains
	 * one entry for every column of the original matrix, where each entry 
	 * contains all row indices with a 'true' values for the corresponding
	 * column. 
	 * 
	 * @return sparse coverage matrix
	 * @param uuts
	 *            rows of the matrices
	 * @param tests
	 *            columns of the matrices
	 * @param testCoverage
	 *            maps each test case (column) to all UUTs it covers
	 * @param includeAllMUTs
	 *            indicates whether MUT that are not of interest (e.g. already
	 *            reviewed) shall be ignored or not
	 */
	@SuppressWarnings("unchecked")
	private static List<Integer>[] getCoverageMatrix( boolean[][] matrix ) {

		if ( matrix.length == 0 )
			return new ArrayList[0];
		
		List<Integer>[] sparseCoverageMatrix = new ArrayList[matrix[0].length];

		for ( int colIndex = 0; colIndex < matrix[0].length; ++colIndex ) {			
			List<Integer> sparseMatrixROW = new ArrayList<Integer>();

			for ( int rowIndex = 0; rowIndex < matrix.length; ++rowIndex )
				if ( matrix[rowIndex][colIndex] == true )
					sparseMatrixROW.add(rowIndex);

			sparseCoverageMatrix[colIndex] = sparseMatrixROW;
		}

		return sparseCoverageMatrix;
	}
}
