package org.ezunit.evaluation;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.math3.util.Precision;
import org.ezunit.evaluation.MatrixDataReader;
import org.ezunit.evaluation.MatrixDataReader.Result;
import org.ezunit.evaluation.MatrixDataReader.Test;
import org.ezunit.evaluation.MatrixDataReader.UUT;
import org.ezunit.evaluation.MatrixDataReader.UUTEntry;

/**
 * Implements the second version of Jones' algorithm from
 * &#064inproceedings{jones2007debugging,
 *  title={Debugging in parallel},
 *  author={Jones, James A and Bowring, James F and Harrold, Mary Jean},
 *  booktitle={Proceedings of the 2007 international symposium on Software testing and analysis},
 *  pages={16--26},
 *  year={2007},
 *  organization={ACM}
 * }
 * by using the data structures from {@link MatrixDataReader}.
 * 
 * @author Wolfgang Hgerle
 */
public final class Jones {
	
	// used to compare two doubles
	private static final double PRECISION = 1E-10;
	
	private static class UUTEntryTarantula implements Comparable<UUTEntryTarantula> {

		public final double tValue;
		public final UUT uut;
		
		public UUTEntryTarantula(UUT uut, double tValue) {
			this.tValue = tValue;
			this.uut = uut;
		}

		@Override
		public int compareTo(UUTEntryTarantula other) {
			// sort descending
			return -Double.compare(this.tValue, other.tValue);
		}
	}
	
	private double mostSuspicious = 0.2; // as suggested in the original paper
	private double similarity = 0.5; // as suggested in the original paper

	private Result result;
	private final Map<Test, List<UUTEntryTarantula>> tarantulaValuesPerSuite = new HashMap<Test, List<UUTEntryTarantula>>();
	private final Map<Test, Set<UUT>> mostSuspiciousSetPerFailedTest = new HashMap<Test, Set<UUT>>();
	private final List<Test> failedTests = new ArrayList<Test>();
	private final Map<Test, Set<Test>> blocksPerTest = new HashMap<Test, Set<Test>>();
	
	/**
	 * Performs the algorithm.
	 * Configures the algorithm in such a way, that it uses the specified parameters. If not called a standard parameter set of 0.2, 0.5 is used.
	 * 
	 * @return	a partition of the Set of Test which served as input.
	 * @param mostSuspicious	A parameter which determines the size of a suspiciousness set.
	 * @param similarity	A threshold for the similarity of two suspiciousness sets. If above the corresponding blocks are joined.
	 */
	public Set<Set<Test>> getBlocks(Result result, double mostSuspicious, double similarity) {
		this.mostSuspicious = mostSuspicious;
		this.similarity = similarity;

		return getBlocks( result );
	}
	
	/**
	 * Performs the algorithm with the set parameter (defaults to 0.2, 0.5).
	 * 
	 * @return	a partition of the Set of Test which served as input.
	 */
	public Set<Set<Test>> getBlocks(Result result) {
		this.result = result;
		
		// first step: construct temporary test suites and calculate suspiciousness for each uut
		split();
		// second step: calculate the set of most suspicious uuts for each test suite
		computeMostSuspiciousSets(true);
		// third step: use these sets to build the blocks
		buildPartition();
		Set<Set<Test>> partition = new HashSet<Set<Test>>();
		partition.addAll(blocksPerTest.values());
		return partition;
	}


	private void buildPartition(){
		// for each test its corresponding block is saved 
		for (Test t : failedTests) {
			Set<Test> p = new HashSet<Test>();
			p.add(t);
			blocksPerTest.put(t, p);
		}

		// now we build the blocks
		for (Test t1 : failedTests) {
			for (int i = failedTests.indexOf(t1)+1; i < failedTests.size(); i++) {
				Test t2 = failedTests.get(i);
				Set<Test> block = blocksPerTest.get(t1);
				if (block.contains(t2)) {
					continue; // t2 ist bereits in der Partition von t1 enthalten und umgekehrt
				}

				double jValue = jaccard(mostSuspiciousSetPerFailedTest.get(t1), mostSuspiciousSetPerFailedTest.get(t2));
				// if two tests are similar
				if (compare(jValue, similarity)>=0) {
					// their corresponding blocks are joined
					block.addAll(blocksPerTest.get(t2));
					for (Test t: block) {
						// we have to keep track
						blocksPerTest.put(t, block);
					}
				}
			}
		}
	}
	
		
	/*
	 *  Calculates sets of most suspicious uuts. 
	 *  @param allTies if there are ties with respect to suspiciousness, should all uuts be considered
	 */
	private void computeMostSuspiciousSets(boolean allTies) {
		
		// there's one test suite for each test
		// each of these test suites gives one set of most suspicious uuts
		for (Test t : failedTests) {
			List<UUTEntryTarantula> allEntries = tarantulaValuesPerSuite.get(t);
			// sort descending by suspiciousness
			Collections.sort(allEntries);
			Set<UUT> mostSuspiciousSet = new HashSet<UUT>();
			// we are interested in the first mostSuspicious * 100 % entries
			int max = (int) Math.round(allEntries.size() * mostSuspicious) - 1; 
			// if only the first entry was affected, max would have become -1
			max = Math.max(max, 0); 
			if (allTies) {
				double minTValue = allEntries.get(max).tValue;
				for (UUTEntryTarantula entry : allEntries) {
					if (compare(entry.tValue,  minTValue) >= 0) {
						mostSuspiciousSet.add(entry.uut);
					}
				}
			} else {
				for (int i = 0; i <= max; i++) {
					mostSuspiciousSet.add(allEntries.get(i).uut);
				}
			}
			// save the resulting set
			mostSuspiciousSetPerFailedTest.put(t, mostSuspiciousSet);
		}
	}

		
	/* 
	 * Build one temporary test suite for each failed test and calculate Tarantula-values for each uut in each test suite.
	 */
	private void split() {
		
		// identify failed and passed tests
		List<Test> passedTests = new ArrayList<Test>();
		for (Test t : result.tests) {
			if (t.passed) {
				passedTests.add(t);
			} else {
				failedTests.add(t);
			}
		}
 
		// count how many passed tests are called by a specific uut
		Map<UUT, Integer> numberOfPassedTestsCalledByUUT = new HashMap<UUT, Integer>();
		Set<UUT> calledUUTsByPassedTests = new HashSet<UUT>();
		for (UUT uut : result.uuts) {
			numberOfPassedTestsCalledByUUT.put(uut, 0);
		}
		for (Test calledTest : passedTests) {
			for (UUTEntry entry : calledTest.calledUUTs) {
				calledUUTsByPassedTests.add(entry.uut);
				numberOfPassedTestsCalledByUUT.put(entry.uut, numberOfPassedTestsCalledByUUT.get(entry.uut) + 1);
			}
		}
		for (Test t : failedTests) {
			// each test suite consists of one failed and all passed tests
			Set<UUT> calledUUTs = new HashSet<UUT>();
			calledUUTs.addAll(calledUUTsByPassedTests);
			for (UUTEntry entry : t.calledUUTs) {
				calledUUTs.add(entry.uut);
			}
			
			// calculate Tarantula-values for each uut and save them for each test suite
			double tValue, p, q;
			List<UUTEntryTarantula> list = new ArrayList<UUTEntryTarantula>();
			for (UUT uut : calledUUTs) {
				
				int failedTotal = 1;
				int passedTotal = passedTests.size();
				
				int failed = 0;
				int passed = numberOfPassedTestsCalledByUUT.get(uut);
				
				if (uut.callingTests.contains(t)) {
					failed = 1;
				}

				if (passedTotal == 0) {
					tValue = 1;
				} else {
					q = ((double) failed) / failedTotal;
					p = ((double) passed) / passedTotal;
					tValue = q/(p+q);
				}
				list.add(new UUTEntryTarantula(uut, tValue));
			}
			tarantulaValuesPerSuite.put(t, list);
		}
	}
	
	
	/*
	 * Calculates Jaccard's set similarity coefficient.
	 * 
	 * @param a	the first set
	 * @param b	the second set
	 * @return	Jaccard's set similarity coefficient
	 */
	private static <T> double jaccard (Set<T> a, Set<T> b){
		int numberOfCommonElements = 0;
		for (T e : a) {
			if (b.contains(e)){
				numberOfCommonElements++;
			}
		}
		
		return ((double) numberOfCommonElements/(a.size() + b.size() - numberOfCommonElements));
	}
	
	/*
	 * Compares two double values with respect to the configured absolute precision {@link #PRECISION}
	 * 
	 * @param x	the first double value
	 * @param y	the second double value
	 * @return 1 if x > y, -1 if x < y and 0 else 
	 */
	private static int compare(double x, double y) {
		return Precision.compareTo(x, y, PRECISION);
	}
}
