package org.ezunit.evaluation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ezunit.evaluation.MatrixDataReader.Result;
import org.ezunit.evaluation.MatrixDataReader.Test;
import org.ezunit.evaluation.MatrixDataReader.UUT;
import org.ezunit.evaluation.MatrixDataReader.UUTEntry;

import com.google.common.collect.Sets;

/**
 * Contains a simple approximation of the maximum set packing problem.
 * This approximation is used to calculate a set of failed tests with disjoint traces. 
 * 
 * @author Wolfgang Hgerle
 */
public final class SubtractiveSetPackingApproximization {
	
	/**
	 * Calculates a set of failed tests from a {@link result} whose traces are disjoint.
	 * Uses {@link #packing(List)} to achieve this goal.
	 * 
	 * @param result the result of a test suite run
	 * @return a set of failed tests with disjoint traces
	 */
	public static Set<Test> getBlocks(Result result) {
		Map<Set<UUT>, Test> testsByUUTs = new HashMap<Set<UUT>, Test>();		
		ArrayList<Set<UUT>> sets = new ArrayList<Set<UUT>>();
		// find the called uuts of each failed test
		for (Test test : result.tests) {
			if (!test.passed) {
				Set<UUT> set = Sets.newIdentityHashSet();
				for (UUTEntry entry : test.calledUUTs) {
					set.add(entry.uut);
				}
				sets.add(set);
				testsByUUTs.put(set, test);
			}
		}
		
		Set<Set<UUT>> packing = packing(sets);
		Set<Test> tests = new HashSet<Test>();
		// now the sets of uuts have to be mapped back to their traces
		for (Set<UUT> packedSet : packing) {
			Test test = testsByUUTs.get(packedSet);
			tests.add(test);
		}
		return tests;
	}

	/**
	 * Calculates an approximation of the maximal set packing problem by a greedy heuristic.
	 * Removes sets from the input as long as the remaining sets constitute no packing.
	 * The idea is to remove first the set overlapping with the greatest number of other sets.
	 * If there are ties, the set which shares most elements with any other present set, is removed.
	 * If there are still ties, sets with more elements are removed first.
	 *  
	 * @param input a list of sets of which a packing is to be calculated 
	 * @return a set-packing as as set of sets
	 */
	public static <E> Set<Set<E>> packing(List<Set<E>> input) {
		Set<Set<E>> result = new HashSet<Set<E>>();
		ArrayList<ElementSet<E>> elementSets = new ArrayList<ElementSet<E>>();
		// construct a list of the input sets with additional information
		for(Set<E> s : input) {
			ElementSet<E> es = new ElementSet<E>(s);
			for(ElementSet<E> other : elementSets) {
				if (!Collections.disjoint(es.getSet(), other.getSet())){
					es.addIntersectingElementSet(other);
					other.addIntersectingElementSet(es);
				}
			}
			elementSets.add(es);
		}
			
		// as long as not all sets are disjoint
		while (!check(elementSets)) {
			// find the set with intersects with most other sets
			ElementSet<E> next = findMostIntersectingElement(elementSets);
			for (ElementSet<E> other : next.getIntersectingElements()) {
				other.removeIntersectingElementSet(next);
			}
			// and remove it
			elementSets.remove(next);
		}
		
		// there are only disjoint sets left
		for (ElementSet<E> es : elementSets) {
			result.add(es.getSet());
		}
		return result;
	}
	
	// Represents a set with additional meta information 
	private static final class ElementSet<E> implements Comparable<ElementSet<E>>{
		private final Set<E> set;
		// saves with how which other sets this set has how many common elements
		private final Map<ElementSet<E>, Integer> intersectingElements = new HashMap<ElementSet<E>, Integer>();
		private final int size;
		// saves the maximum number of elements this set has in common with any single other set
		private int max = 0;
		
		public ElementSet(Set<E> set) {
			this.set = set;
			this.size = set.size();
		}
		
		public void addIntersectingElementSet(ElementSet<E> intersectingSet) {
			// we have to update some information if we add a new possibly intersecting set
			int numberOfCommonElements = Sets.intersection(this.set, intersectingSet.getSet()).size();
			intersectingElements.put(intersectingSet, numberOfCommonElements);
			if(numberOfCommonElements > max) {
				max = numberOfCommonElements;
			}
		}
		
		@Override
		public int compareTo(ElementSet<E> other) {
			// a set is "bigger" than another set if it overlaps with a greater number of other sets
			if (this.getNumberOfIntersectingElements() > other.getNumberOfIntersectingElements()) {
				return 1;
			}
			if (this.getNumberOfIntersectingElements() < other.getNumberOfIntersectingElements()) {
				return -1;
			}
			// in case of a tie, number of common elements counts
			if (this.getMax() > other.getMax()) {
				return 1;
			}
			if (this.getMax() < other.getMax()) {
				return -1;
			}
			// If there is still a tie, sets with smaller numbers of elements are preferred.
			// Leads to packings whose sets are preferably small.
			if (this.getSize() > other.getSize()) {
				return 1;
			}
			if (this.getSize() < other.getSize()) {
				return -1;
			}
			return 0;
		}
		
		public Set<ElementSet<E>> getIntersectingElements() {
			return intersectingElements.keySet();
		}
		
		public int getMax() {
			return max;
		}
		
		public int getNumberOfIntersectingElements() {
			return intersectingElements.size();
		}
		
		public Set<E> getSet() {
			return set;
		}
		
		public int getSize() {
			return size;
		}
		
		public void removeIntersectingElementSet(ElementSet<E> intersectingSet) {
			// if one removes an intersecting set, one has to update some meta information
			int value = intersectingElements.remove(intersectingSet);
			if (value >= max) {
				findMax();
			}
		}

		@Override
		public String toString() {
			StringBuilder result = new StringBuilder();
			result.append("Size " + this.getSize() + "\n");
			result.append("Intersects with " + this.getNumberOfIntersectingElements() + " elements\n");
			result.append("Elements\n");
			for (E e : this.set) {
				result.append(e.toString()+"\n");
			}
			return result.toString();
		}
		
		private void findMax() {
			// we have saved the number of common elements as values in a map 
			Collection<Integer> values = intersectingElements.values();
			if(!values.isEmpty()) {
				this.max = Collections.max(values);
			}
		}
	}

	// checks if all sets of the input list are disjoint
	private static <E> boolean check(List<ElementSet<E>> elementSets) {
		boolean result = true;
		for (ElementSet<E> es : elementSets) {
			if(es.getNumberOfIntersectingElements()>0) {
				result = false;
			}
		}
		return result;
	}
	
	// find the "biggest" set according to the predefined order
	private static <E> ElementSet<E> findMostIntersectingElement(List<ElementSet<E>> elementSets) {
		ElementSet<E> result = elementSets.get(0);
		if (result != null) {
			for (ElementSet<E> es : elementSets) {
				if(es.compareTo(result) > 0 ) {
					result = es;
				}
			}
		}
		return result;
	}
}