package org.ezunit.evaluation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ezunit.evaluation.MatrixDataReader.Element;
import org.ezunit.evaluation.MatrixDataReader.Result;
import org.ezunit.evaluation.MatrixDataReader.Test;
import org.ezunit.evaluation.MatrixDataReader.UUT;
import org.ezunit.evaluation.MatrixDataReader.UUTEntry;

class Matrix<RowType extends Element, ColType extends Element> {
	
	// attributes
	public RowType[] rowValues;
	public ColType[] colValues;
	
	public boolean[][] 	matrix;			// full matrix
	
	// constructor
	public Matrix( RowType[] rows, ColType[] cols ) {
		rowValues = rows.clone();
		colValues = cols.clone();
		
		matrix = new boolean[rows.length][cols.length];
		
		// build the matrix
		for ( int row = rows.length - 1; row >= 0; --row )
			for ( int col = cols.length - 1; col >= 0; --col )
				if ( rows[row].covers( cols[col] ) )
					matrix[row][col] = true;
	}
}

class WKBlock<RowType, ColType> {
	// attributes
	public List<RowType> 		dirt			= new ArrayList<RowType>();			// dirt of this block
	
	public List<List<RowType>> 	subBlocks		= new ArrayList<List<RowType>>();	// sub blocks           
	
	public List<ColType> 		cols			= new ArrayList<ColType>();			// columns of this block	
}

/**
 * Represents a block.
 * The elements of the block are indices into a corresponding matrix which 
 * has to be given separately and which is used during block building. 
 */
class Block {
	// attributes
	public List<Integer>		dirt			= new ArrayList<Integer>();		// dirt of this block
	public List<Integer>		remainingRows	= new ArrayList<Integer>();		// remaining rows not covered by dirt or sub blocks (always empty if sub blocks exist)
	
	public List<Integer>		cols			= new ArrayList<Integer>();		// columns of this block (includes columns of sub-blocks)
	
	public LinkedList<Block>	subblocks		= new LinkedList<Block>();		// sub blocks
	
	// constructors
	
	/**
	 * Empty constructor.
	 */
	public Block() {
	}
	
	/**
	 * Copy constructor.
	 * Copies all elements of the base block, but for the sub blocks
	 * 
	 * @param base
	 */
	public Block( Block base ) {
		dirt.addAll( base.dirt );
		remainingRows.addAll( base.remainingRows );
		cols.addAll( base.cols );
	}
}

class Partition<RowType extends Element, ColType extends Element> {
	// attributes
	public Matrix<RowType, ColType> matrix;
	public Block					block;
	
	// constructor
	public Partition( Matrix<RowType, ColType> matrix, Block block ) {
		this.matrix = matrix;
		this.block = block;
	}
	
	// methods
	@Override
	public String toString() {
		return "Partition (" + matrix.rowValues.length + ", " + matrix.colValues.length + ")";
	}
}

class Pair<F,S> {
	public F first; 
	public S second;
	
	public Pair( F first, S second ) { 
		this.first = first; 
		this.second = second; 
	}	
	
	@Override
	public int hashCode() { 
		return 31 * first.hashCode() + ( second != null ? second.hashCode() : 0 );
	}
	
	@Override
	public boolean equals( Object o ) {
		return o instanceof Pair<?,?> && first.hashCode() == ( (Pair<?,?>)o ).first.hashCode() && second.hashCode() == ( (Pair<?,?>)o ).second.hashCode(); 
	}
}

final class NoBlockAngularFormException extends Exception {
	private static final long	serialVersionUID	= 4437115584888252882L;
}

final class RowColPair { 
	public RowColPair( int row, int col ) { 
		this.row = row; this.col = col; 
	} 
	
	public int row; 
	public int col; 
}

final class IndexBlockArrayList {
	
	private IndexBlock parent;
	public int[] data = new int[0];
	
	public IndexBlockArrayList( IndexBlock parent ) {
		super();
		this.parent = parent;
		
		assert( parent != null );
	}

	public IndexBlockArrayList( IndexBlock parent, IndexBlockArrayList base ) {
		this.parent = parent;
		this.data = base.data.clone();
		
		assert( parent != null );
	}
	
	public IndexBlockArrayList( IndexBlock parent, ArrayList<Integer> base ) {
		this.parent = parent;
		this.data = new int[base.size()];
		
		int index = 0;
		
		for ( Integer i : base )
			data[index++] = i;
		
		assert( parent != null );
	}

	public void add( int value ) {
		parent.invalidateHash();
		
		int oldLength = data.length;
		
		data = Arrays.copyOf( data, oldLength + 1 );
		data[oldLength] = value; 
	}
	
	public void addAll( IndexBlockArrayList c ) {
		parent.invalidateHash();
		
		int oldLength = data.length;
		
		data = Arrays.copyOf( data, oldLength + c.data.length );
		
		for ( int i = 0; i < c.data.length; ++i )
			data[oldLength++] = c.data[i]; 
	}
	
	public void addAll( List<Integer> c ) {
		parent.invalidateHash();
		
		int oldLength = data.length;
		
		data = Arrays.copyOf( data, oldLength + c.size() );
		
		for ( Integer i : c )
			data[oldLength++] = i;
	}

	public void removeAll( List<Integer> rowsToRemove ) {
    	parent.invalidateHash();
    	
		LinkedList<Integer> elementsToKeep = new LinkedList<Integer>();
		
		for ( int i : data )
			if ( !rowsToRemove.contains( i ) )
				elementsToKeep.add( i );
		
		data = new int[elementsToKeep.size()];
		
		int index = 0;
		
		for ( Integer e : elementsToKeep )  
            data[index++] = e;
	}
	
	public int size() {
		return data.length;
	}

	public boolean isEmpty() {
		return data.length == 0;
	}

	public void sort() {
		parent.invalidateHash();
		Arrays.sort( data );
	}

	public boolean contains( int value ) {
		for ( int i : data )
			if ( i == value )
				return true;
		return false;
	}
	
	@Override
	public int hashCode() {
		int hashCode = 1;
		
		for ( int i : data )
			hashCode = 31 * hashCode + i;
		
		return hashCode;
	}
}    

final class IndexBlock {
	IndexBlockArrayList rows = new IndexBlockArrayList( this );
	IndexBlockArrayList cols = new IndexBlockArrayList( this );

	// Access this field only if you're absolutely knowing what you're doing
	// (that means, if you're sure, that the hash has been validated before usage)
	// and NEVER write the value directly!
	// It is only made protected for performance reasons.
	protected int		hash			= 0;

	private boolean		hashComputed	= false;

	public void invalidateHash() {
		hashComputed = false;
	}

	public void validateHash() {
		internHashCode();
	}

	@Override
	public IndexBlock clone() {
		IndexBlock block = new IndexBlock();

		block.rows = new IndexBlockArrayList( block, rows );
		block.cols = new IndexBlockArrayList( block, cols );

		return block;
	}

	@Override
	public String toString() {
		String str = "({";

		for ( int i : rows.data )
			str += i + ",";

		str = str.substring( 0, str.length() - 1 );
		str += "}, {";

		for ( int i : cols.data )
			str += i + ",";

		str = str.substring( 0, str.length() - 1 );
		str += "})\n";

		return str;
	}

	@Override
	public boolean equals( Object o ) {
		if ( o instanceof IndexBlock )
			return internHashCode() == ( (IndexBlock)o ).internHashCode();

		return false;
	}
	
	public boolean equalsIndexBlock( IndexBlock o ) {
		return internHashCode() == o.internHashCode();
	}

	private int internHashCode() {
		if ( !hashComputed ) {
			hash = ( rows.hashCode() << 16 ) + cols.hashCode();
			hashComputed = true;
		}

		return hash;
	}
}

final class OptimizedIndexBlock {
	public IndexBlock		dirt		= new IndexBlock();
	public List<IndexBlock>	subblocks	= new LinkedList<IndexBlock>();

	public double			rowThreshold;
	public double			columnThreshold;
	public double			coverageThreshold;
}

final class MatchingResult {
	LinkedList<HashSet<IndexBlock>> blockSetsToMerge 		= new LinkedList<HashSet<IndexBlock>>();
	LinkedList<Pair<Integer, IndexBlock>> indicesToMerge 	= new LinkedList<Pair<Integer,IndexBlock>>();	// interpretation of indices as row or column possible  
	LinkedList<Integer> dirtIndices 						= new LinkedList<Integer>();					// interpretation of indices as row or column possible
	
	double threshold = 0;
	
	@Override
	public int hashCode() {
		return ( (Double)threshold ).hashCode();
	}
	
	@Override
	public boolean equals( Object o ) {
		if ( o instanceof MatchingResult ) {
			MatchingResult second = (MatchingResult)o;
			
			if ( blockSetsToMerge.hashCode() == second.blockSetsToMerge.hashCode() 
					&& indicesToMerge.hashCode() == second.indicesToMerge.hashCode()
					&& dirtIndices.hashCode() == second.dirtIndices.hashCode() )
				return true;
		}
		
		return false;
	}
	
	public boolean equalsMatchingResult( MatchingResult second ) {
		if ( blockSetsToMerge.hashCode() == second.blockSetsToMerge.hashCode() 
			&& indicesToMerge.hashCode() == second.indicesToMerge.hashCode()
			&& dirtIndices.hashCode() == second.dirtIndices.hashCode() )
				return true;
		return false;
	}
}

final class MatchingSearchRange {
	double startRange;
	double endRange;
	
	MatchingResult startMatching;
	MatchingResult endMatching;
	
	int depth;
}

interface IMatchingFunctor {
	public MatchingResult call( boolean[][] matrix, 
		LinkedList<IndexBlock> result, 
		List<Integer> unmatchedIndices, 
		double coverageThreshold );
}

/**
 * Builds the blocks from the given matrices.
 * For each matrix, a (possibly empty) block will be generated.
 *
 * @param <RowType>
 * @param <ColType>
 */
public class WK {
	private final int maxSteps = 30;
	
	// methods
	
	/**
	 * Initiates the block building
	 */
	public Set<WKBlock<UUT, Test>> getBlocksH( Result input ) {
		// get all partitions
		Set<Set<Test>> parts = BDM.getBlocks( input );
		
		// extract dirt for every partition
		Set<WKBlock<UUT, Test>> result = new HashSet<WKBlock<UUT, Test>>();
		
		for ( Set<Test> part : parts ) {
			// gather all tests and UUTs
			Set<UUT> uuts = new HashSet<UUT>();
			
			for ( Test test : part )
				for ( UUTEntry entry : test.calledUUTs )
					uuts.add( entry.uut );
			
			result.add( getBlock( uuts.toArray( new UUT[uuts.size()] ), part.toArray( new Test[part.size()] ) ) );
		}
		
		return result;
	}
	
	public Set<WKBlock<Test, UUT>> getBlocksV( Result input ) {
		// get all partitions
		Set<Set<Test>> parts = BDM.getBlocks( input );
		
		// extract dirt for every partition
		Set<WKBlock<Test, UUT>> result = new HashSet<WKBlock<Test, UUT>>();
		
		for ( Set<Test> part : parts ) {
			// gather all tests and UUTs
			Set<UUT> uuts = new HashSet<UUT>();
			
			for ( Test test : part )
				for ( UUTEntry entry : test.calledUUTs )
					uuts.add( entry.uut );
			
			result.add( getBlock( part.toArray( new Test[part.size()] ), uuts.toArray( new UUT[uuts.size()] ) ) );
		}
		
		return result;
	}
	
	private <RowType extends Element, ColType extends Element> WKBlock<RowType, ColType> getBlock( RowType[] rows, ColType[] cols ) {
		// create the matrix
		Matrix<RowType, ColType> matrix = new Matrix<RowType, ColType>( rows, cols );
		
		// create the index arrays
		Integer[] rowsIndices = new Integer[rows.length];
		Integer[] colsIndices = new Integer[cols.length];
		
		for ( int i = rows.length - 1; i >= 0; --i )
			rowsIndices[i] = i;

		for ( int i = cols.length - 1; i >= 0; --i )
			colsIndices[i] = i;
		
		Block b = extractBlocks( rowsIndices, colsIndices, matrix );
		
		// create the final block
		WKBlock<RowType, ColType> block = new WKBlock<RowType, ColType>();

		for ( Integer rowIndex : b.dirt )
			block.dirt.add( rows[rowIndex] );
		
		for ( Integer rowIndex : b.remainingRows )
			block.dirt.add( rows[rowIndex] );
		
		for ( Block bb : b.subblocks ) {
			LinkedList<RowType> subblock = new LinkedList<RowType>();

			for ( Integer rowIndex : bb.dirt )
				subblock.add( rows[rowIndex] );
			
			for ( Integer rowIndex : bb.remainingRows )
				subblock.add( rows[rowIndex] );
			
			block.subBlocks.add( subblock );
		}

		for ( Integer colIndex : b.cols )
			block.cols.add( cols[colIndex] );
		
		return block;
	}
	
	// private methods
	
	/**
	 * Extracts the blocks (as sub blocks of the return value) from the given matrix. 
	 * Tries to find the blocks in a way such that the amount is minimal, but above zero.
	 * @param rowIndices indices of the rows to use
	 * @param colIndices indices of the columns to use
	 * @param matrix original matrix
	 * @return
	 */
	private <RowType extends Element, ColType extends Element> Block extractBlocks( Integer[] rowIndices, Integer[] colIndices, Matrix<RowType, ColType> matrix ) {
		// count the number of covered rows/columns
		final RowType[] rows = matrix.rowValues;
		final ColType[] cols = matrix.colValues;
		
		final Map<RowType, Integer> rowCount = new HashMap<RowType, Integer>();
		final Map<ColType, Integer> colCount = new HashMap<ColType, Integer>();

		// count rows and columns; while doing so, also filter out 
		// all rows which either cover all columns or are 
		// touched by a column which covers only that row 
		// note: DON't filter out full coverage and single coverage rows
		//       in advance, because that will mess up further dirt 
		//       recognition
		Set<Integer> filteredRows = new HashSet<Integer>();
		
		// count rows
		for ( Integer row : rowIndices ) {
			int count = 0;
			
			for ( Integer col : colIndices )
				if ( matrix.matrix[row][col] == true )
					++count;
			
			rowCount.put( rows[row], count );
		}
		
		// count columns
		for ( Integer col : colIndices ) {
			int count = 0;
			
			for ( Integer row : rowIndices )
				if ( matrix.matrix[row][col] == true ) {
					++count;
				}
			
			colCount.put( cols[col], count );
		}
		
		// prepare the row and column indices arrays
		Integer[] workingRowIndices = new Integer[rowIndices.length - filteredRows.size()];
		Integer[] workingColIndices = colIndices.clone();
		
		int index = 0;

		for ( Integer row : rowIndices )
			if ( !filteredRows.contains( row ) )
				workingRowIndices[index++] = row;
		
		// order the rows by their column coverage
		Arrays.sort( workingRowIndices, new Comparator<Integer>() {
			public int compare( Integer o1, Integer o2 ) {
				int i1 = rowCount.get( rows[o1] );
				int i2 = rowCount.get( rows[o2] );
				
				return i1 > i2 ? -1 : i1 == i2 ? rows[o1].getName().compareTo( rows[o2].getName() ) : 1;
			}});
		
		// order the columns by their row coverage
		Arrays.sort( workingColIndices, new Comparator<Integer>() {
			public int compare( Integer o1, Integer o2 ) {
				int i1 = colCount.get( cols[o1] );
				int i2 = colCount.get( cols[o2] );

				return i1 > i2 ? -1 : i1 == i2 ? cols[o1].getName().compareTo( cols[o2].getName() ) : 1;
			}});
		
		// get the result
		Block result = null;
		
		// create an empty block if no rows are left
		if ( workingRowIndices.length == 0 ) {
			result = new Block();
			result.cols.addAll( Arrays.asList( colIndices ) );
		}
		else
			result = extractBlocksHelper( workingRowIndices, workingColIndices, matrix );

		if ( result != null ) {
			// if there are no special rows and only one block in the result, pull this block up 
			if ( filteredRows.size() == 0 && result.dirt.size() == 0 && result.subblocks.size() == 1 )
				result = new Block( result.subblocks.get( 0 ) );
			
			// add the previously filtered rows
			for ( Integer row : filteredRows )
				result.dirt.add( 0, row );
		}
		
		return result;
	}
	
	/**
	 * Does the actual work for the block extraction
	 * @param rowIndices indices of the rows to use
	 * @param colIndices indices of the columns to use
	 * @param matrix original matrix
	 * @return
	 */
	private <RowType extends Element, ColType extends Element> Block extractBlocksHelper( Integer[] rowIndices, Integer[] colIndices, Matrix<RowType, ColType> matrix ) {
		// build the sub matrix from the given rows and columns to allow the block finding algorithm to work on that
		// (will save index lookups into the rowIndices and colIndices arrays)
		boolean[][] subMatrix = new boolean[rowIndices.length][colIndices.length];
		
		for ( int row = rowIndices.length - 1; row >= 0; --row )
			for ( int col = colIndices.length - 1; col >= 0; --col )
				subMatrix[row][col] = matrix.matrix[rowIndices[row]][colIndices[col]];
		
		// calculate the indices of the blocks
		// note: the indices will be indices into the subMatrix (i.e. in sub matrix space), not the original matrix
		OptimizedIndexBlock blockList; 

		try {
			LinkedList<RowColPair> matching = calcMaximumMatching( subMatrix );   
			List<IndexBlock> blocks			= getBlocks( subMatrix, matching );
			blockList 						= optimizeBlocks( subMatrix, blocks );
		} catch( NoBlockAngularFormException e ) {
			// matrix has no block angular form - not much we can do about that
			e.printStackTrace();
			
			blockList = new OptimizedIndexBlock();
		}
		
		// create the top level block for this set of rows
		Block topLevelBlock = new Block();

		// add the dirt		
		for ( int row : blockList.dirt.rows.data )
			topLevelBlock.dirt.add( rowIndices[row] );
		
		for ( IndexBlock block : blockList.subblocks ) {
			Block pb = new Block();
			
			for ( int row : block.rows.data )
				pb.remainingRows.add( rowIndices[row] );
			
			for ( int col : block.cols.data )
				pb.cols.add( colIndices[col] );
		
			topLevelBlock.subblocks.add( pb );
		}
		
		// sort the blocks by their size
		Collections.sort( topLevelBlock.subblocks, new Comparator<Block>() {
			public int compare( Block o1, Block o2 ) {
				return o1.cols.size() - o2.cols.size();
			}});

		// add the remaining tests
		topLevelBlock.cols.addAll( Arrays.asList( colIndices ) );

		return topLevelBlock;
	}
	
	/**
	 * Get the ordered blocks of the inscribed square matrix of matrix described by the given
	 * row-column pairs. Blocks are always squared at this point.
	 * @param matrix matrix to operate on
	 * @param matching the given matching
	 * @return list of blocks
	 * @throws Exception 
	 */
	private List<IndexBlock> getBlocks( boolean[][] baseMatrix, LinkedList<RowColPair> pairs ) throws NoBlockAngularFormException {
		// clone the matrix because it has to be altered
		boolean[][] matrix = baseMatrix.clone();
		
		// the row and column indices to use
		Integer[] rowIndices = new Integer[pairs.size()]; 
		Integer[] colIndices = new Integer[pairs.size()]; 
		
		int pairIndex = 0;
		
		for ( RowColPair p : pairs ) {
			rowIndices[pairIndex] = p.row;
			colIndices[pairIndex++] = p.col;
		}
		
		int rowsLeft = pairs.size();
		int bitCount;
		int row, col;
		
		// contains the indices of all blocks into the pairs array
		List<List<Integer>> blockIndices = new ArrayList<List<Integer>>();
		
		ArrayList<Integer> rowChain = new ArrayList<Integer>();

		// associates a row index with a created block
		Map<Integer, List<Integer>> associatedBlocks = new HashMap<Integer, List<Integer>>();
	
		boolean foundRows = false;
		
		// do the work
		while ( rowsLeft > 0 ) {		
			// identify all rows with no off-diagonal bits
			// since found rows may lead to new rows without off-diagonal bits, search until no new row is found!
			do {
				foundRows = false;
				
				for ( int rowIndex = 0; rowIndex < rowIndices.length; ++rowIndex ) {
					if ( rowIndices[rowIndex] == null ) continue;	// skip ignored rows
	
					bitCount = 0;
					row = rowIndices[rowIndex];
					
					for ( int colIndex = 0; colIndex < colIndices.length; ++colIndex ) {
						if ( colIndices[colIndex] == null ) continue;	// skip ignored columns
						
						if ( matrix[row][colIndices[colIndex]] == true )
							++bitCount;	
						
						if ( bitCount > 1 )
							break;	// no sense in searching further bits
					}
					
					// if there are no off-diagonal bits
					if ( bitCount == 1 ) { // there is at least one bit in every row (the diagonal bit), which will be ignored
						// create a new block only if the block is not already created (will happen with larger blocks)
						if ( !associatedBlocks.containsKey( rowIndex ) ) {
							List<Integer> block = new ArrayList<Integer>();
							block.add( rowIndex );
							blockIndices.add( block );
							associatedBlocks.put( rowIndex, block );
						} 
						else {
							// important: remove the block from the blocks list and add it anew, moving the block to the back of the list
							// that is important to get the correct order of the blocks
							List<Integer> block = associatedBlocks.get( rowIndex );
							blockIndices.remove( block );
							blockIndices.add( block );
						}
						
						// ignore the row and its corresponding column from now on
						rowIndices[rowIndex] = null;
						colIndices[rowIndex] = null;
						
						--rowsLeft;
						
						foundRows = true;
					}
				}
			} while ( foundRows == true );
			
			// find the first row with an off-diagonal bit 
			outerLoop: for ( int rowIndex = 0; rowIndex < rowIndices.length; ++rowIndex ) {
				if ( rowIndices[rowIndex] == null ) continue;	// skip ignored rows
			
				row = rowIndices[rowIndex];
				
				for ( int colIndex = 0; colIndex < colIndices.length; ++colIndex ) {
					if ( colIndices[colIndex] == null ) continue;	// skip ignored columns
					
					col = colIndices[colIndex];
					
					// found the first off-diagonal bit (use rowIndex and colIndex for comparison, because they form the virtual matrix we are working on)
					if ( rowIndex != colIndex && matrix[row][col] == true ) {
						// construct a row chain to find all associated rows
						rowChain.clear();
						
						rowChain.add( rowIndex );
						
						innerLoop: while ( true ) {
							rowIndex = colIndex;	// search the next row associated with the given column
							
							if ( rowIndices[rowIndex] == null )
								throw new NoBlockAngularFormException();
							
							row = rowIndices[rowIndex];
							rowChain.add( rowIndex );
							
							// find the first off-diagonal bit in the new row
							for ( colIndex = 0; colIndex < colIndices.length; ++colIndex ) {
								if ( colIndices[colIndex] == null ) continue;	// skip ignored columns
					
								col = colIndices[colIndex];
								
								// found the first off-diagonal bit
								if ( rowIndex != colIndex && matrix[row][col] == true ) {
									// if the new row index (yet the colIndex) is not already in the chain, continue
									if ( !rowChain.contains( colIndex ) ) {
										continue innerLoop;
									}
									else {
										// if the row is already in the chain, we found a closed row chain (cycle) and can now create a new block
										
										// delete all rows from the chain that are not within the cycle (the cycle begin is stored in colIndex at this point)
										int numRowsToDelete = rowChain.indexOf( colIndex );
										
										for ( int i = 0; i < numRowsToDelete; ++i )
											rowChain.remove( 0 );
										
										int firstRowIndex = rowChain.get( 0 );

										// create a new block which contains all indices of the found, existing blocks and the remaining rows
										// for this purpose, get all existing blocks that are involved in the chain
										List<Integer> block = new ArrayList<Integer>();
										
										for ( Integer tempRowIndex : rowChain ) {
											// if there is a block already associated with this row, add the whole block to the new block (and remove it from the map/list of blocks)
											if ( associatedBlocks.containsKey( tempRowIndex ) ) {
												block.addAll( associatedBlocks.get( tempRowIndex ) );
												
												blockIndices.remove( associatedBlocks.get( tempRowIndex ) );	// remove the block from the block indices list because it is no longer required
												associatedBlocks.remove( tempRowIndex );						// remove the block from the map (same reason)
											}
											else
												block.add( tempRowIndex );	// if there is no block associated with this row, just add the row to the new block
										}

										// associate the first row of the new block with the block (because all other rows will be ignored in the next step)
										associatedBlocks.put( block.get( 0 ), block );
										
										// store the block
										blockIndices.add( block );
										
										// OR the rows and associated columns of the row chain into the first row of the chain (which can be deleted beforehand from the chain)
										rowChain.remove( 0 );
										
										// adjust the remaining row count
										rowsLeft -= rowChain.size();
										
										int firstRow = rowIndices[firstRowIndex];
										int firstCol = colIndices[firstRowIndex];
										
										for ( Integer tempRowIndex : rowChain ) {
											int tempRow = rowIndices[tempRowIndex];
											
											// process all row-bits
											for ( int i = 0; i < pairs.size(); ++i )
												if ( colIndices[i] != null ) // only consider not-ignored columns
													matrix[firstRow][colIndices[i]] 
													  = matrix[firstRow][colIndices[i]] || matrix[tempRow][colIndices[i]];
											
											int tempCol = colIndices[tempRowIndex];
											
											// process all columns bits
											for ( int i = 0; i < pairs.size(); ++i )
												if ( rowIndices[i] != null ) // only consider not-ignored rows
													matrix[rowIndices[i]][firstCol] 
													  = matrix[rowIndices[i]][firstCol] || matrix[rowIndices[i]][tempCol];
											
											// ignore this row and column from now on
											rowIndices[tempRowIndex] = null;
											colIndices[tempRowIndex] = null;
										}
										
										break innerLoop;
									}
								}
							}
							
							// if this point is reached, the target row had no off-diagonal bits
							throw new NoBlockAngularFormException();
						}
						
						break outerLoop; // continue with the next overall-step
					}
				}
			}
		}
		
		// translate the indices in the block list into real row and column numbers 
		List<IndexBlock> blocks = new ArrayList<IndexBlock>();
		
		for ( List<Integer> block : blockIndices ) {
			IndexBlock newBlock = new IndexBlock();
			
			List<Integer> rowsToAdd = new LinkedList<Integer>();
			List<Integer> colsToAdd = new LinkedList<Integer>();
			
			for ( Integer index : block ) {
				rowsToAdd.add( pairs.get( index ).row );
				colsToAdd.add( pairs.get( index ).col );
			}
			
			newBlock.rows.addAll( rowsToAdd );
			newBlock.cols.addAll( colsToAdd );
			
			blocks.add( newBlock );
		}
		
		return blocks;
	}

	/**
	 * Tries to optimize the given block structure by merging suitable blocks and moving unmatched rows/columns into the block structure.
	 * @param matrix matrix to operate on
	 * @param blocks list of blocks (with indices into matrix) to optimize
	 * @param mergeParams parameter for the merging algorithm
	 * @return optimized block list
	 */
	private OptimizedIndexBlock optimizeBlocks( boolean[][] matrix, List<IndexBlock> blocks ) {
		// clone the blocks and get the matched rows/columns
		LinkedList<IndexBlock> result = new LinkedList<IndexBlock>();
		LinkedList<IndexBlock> blocksToRemove = new LinkedList<IndexBlock>();
		
		List<Integer> matchedRows = new ArrayList<Integer>( matrix.length );
		List<Integer> matchedCols = new ArrayList<Integer>( matrix[0].length );
		
		for ( IndexBlock block : blocks ) {
			result.add( block.clone() );
			
			for ( int row : block.rows.data )
				matchedRows.add( row );

			for ( int col : block.cols.data )
				matchedCols.add( col );
		}
			
		// just for safety
		if ( matrix.length == 0 ) {
			OptimizedIndexBlock list = new OptimizedIndexBlock();
			list.subblocks = result;
			return list;
		}
		
		// get the unmated rows/columns
		List<Integer> unmatchedRows = new ArrayList<Integer>( matrix.length );
		List<Integer> unmatchedCols = new ArrayList<Integer>( matrix[0].length );
		
		for ( int i = 0; i < matrix.length; ++i )
			if ( !matchedRows.contains( i ) )
				unmatchedRows.add( i );

		for ( int i = 0; i < matrix[0].length; ++i )
			if ( !matchedCols.contains( i ) )
				unmatchedCols.add( i );
				
		/* note:
		 	x: verticalBlockCoverageThreshold (row by row)
			y: horizontalBlockCoverageThreshold (column by column)
			z: blockRowCoverageTreshold (bitwise)
		
			step 1: depending on unmatched columns and vertical size of connected blocks
				<x: more dirt, more and smaller blocks
				>x: less dirt, less and larger blocks

			step 2: depending on unmatched rows and horizontal size of connected blocks
				<y: more dirt, more and smaller blocks
				>y: less dirt, less and larger blocks
			
			step 3: depending on block sizes and off-diagonal bits
				<z: less dirt, less and larger blocks
				>z: more dirt, more and smaller blocks
				
				<block sizes: less dirt, less and larger blocks 
				>block sizes: more dirt, more and smaller blocks
		*/
		
		/*
		 	x -> >block size, <#dirt, <#blocks
		 	y -> >block size, <#dirt, <#blocks
		 	
		 	z -> <block size, >#dirt, >#blocks
		 	block size -> <block size, >#dirt, >#blocks
		 */
						
		LinkedList<MatchingResult> columnMatchings;
		LinkedList<MatchingResult> rowMatchings;
		
		// step 1: merge all blocks that are connected by unmatched columns, including all off-diagonal bits that would be part of the new block
		//			(the inclusion of off-diagonal bits will happen automatically during block merging in this implementation);
		//		   merge only blocks where the rows of the blocks connected by a column do not cover more than x% of all rows; else, move all rows covered by the column to the dirt	
		// e.g.				matched	    |	unmatched
		//					1		    |
		//					  2		    |	x x 
		//					  x 3	    |	x x
		//					      4     |	    x x
		//
		// would lead to	matched	    |	unmatched
		//					1
		//					  2   2 2   |
		//					  2 2 2 2   |
		//					          3 |
		
		// calculate all possible column matchings
		columnMatchings = new LinkedList<MatchingResult>();
		columnMatchings.add( matchUnmatchedColumns( matrix, result, unmatchedCols, 0 ) );
						
		// step 2: repeat step 1, but for the unmatched rows; store only blocks where the columns of the blocks connected by a row do not cover more than x% of all columns; else, move the row to the dirt
		
		// calculate all possible row matchings
		rowMatchings = new LinkedList<MatchingResult>();
		rowMatchings.add( matchUnmatchedRows( matrix, result, unmatchedRows, 0 ) );
	
		// iterate over all possible configurations, calculating for each configuration several results to find the best one
		OptimizedIndexBlock bestResult = null;
		
		// sort the matchings by their threshold
		Collections.sort( columnMatchings, new Comparator<MatchingResult>() {
			public int compare( MatchingResult o1, MatchingResult o2 ) {
				return o1.threshold > o2.threshold ? -1 : o1.threshold == o2.threshold ? 0 : 1;
			}} );
		
		Collections.sort( rowMatchings, new Comparator<MatchingResult>() {
			public int compare( MatchingResult o1, MatchingResult o2 ) {
				return o1.threshold > o2.threshold ? -1 : o1.threshold == o2.threshold ? 0 : 1;
			}} );
		
		optimizerLoop: for ( MatchingResult columnMatching : columnMatchings ) {
			for ( MatchingResult rowMatching : rowMatchings ) {
				// clone the initial values to work on a separate set of data
				ArrayList<IndexBlock> matchingResult = new ArrayList<IndexBlock>();
				IndexBlock[] matchingResultArray = new IndexBlock[result.size()];
				
				int index = 0;
				
				for ( IndexBlock block : result ) {
					block = block.clone();
					matchingResult.add( block );
					matchingResultArray[index++] = block;
				}
						
				// get the new blocks to stay in sync with the cloned data set
				LinkedList<Pair<Integer, IndexBlock>> colsToMerge = new LinkedList<Pair<Integer, IndexBlock>>();
				LinkedList<Pair<Integer, IndexBlock>> rowsToMerge = new LinkedList<Pair<Integer, IndexBlock>>();
				LinkedList<HashSet<IndexBlock>> blockSetsToMerge = new LinkedList<HashSet<IndexBlock>>();
						
				for ( Pair<Integer, IndexBlock> pair : columnMatching.indicesToMerge )
					colsToMerge.add( new Pair<Integer, IndexBlock>( pair.first, selectBlock( matchingResultArray, pair.second ) ) );
				
				for ( Pair<Integer, IndexBlock> pair : rowMatching.indicesToMerge )
					rowsToMerge.add( new Pair<Integer, IndexBlock>( pair.first, selectBlock( matchingResultArray, pair.second ) ) );
						
				for ( HashSet<IndexBlock> colSet : columnMatching.blockSetsToMerge ) {
					HashSet<IndexBlock> newSet = new HashSet<IndexBlock>();
					
					for ( IndexBlock block : colSet )
						newSet.add( selectBlock( matchingResultArray, block ) );
							
					blockSetsToMerge.add( newSet );
				}
				
				// merge the merge sets of the row matching into the merge sets of the column matching
				for ( HashSet<IndexBlock> rowSet : rowMatching.blockSetsToMerge ) {
					HashSet<IndexBlock> newSet = new HashSet<IndexBlock>();
					
					for ( IndexBlock block : rowSet )
						newSet.add( selectBlock( matchingResultArray, block ) );
							
					LinkedList<HashSet<IndexBlock>> matchingBlockSets = new LinkedList<HashSet<IndexBlock>>();
					
					for ( HashSet<IndexBlock> set : blockSetsToMerge )
						for ( IndexBlock b : newSet )
							if ( set.contains( b ) ) {
								matchingBlockSets.add( set );
								break;
							}

					if ( matchingBlockSets.size() > 1 ) {
						// merge block sets
						HashSet<IndexBlock> blockSet = new HashSet<IndexBlock>();
						
						for ( HashSet<IndexBlock> set : matchingBlockSets )
							blockSet.addAll( set );
						
						blockSet.addAll( newSet );
						
						blockSetsToMerge.removeAll( matchingBlockSets );
						blockSetsToMerge.add( blockSet );
					}
					else if ( matchingBlockSets.size() == 1 ) {
						// just add the found blocks to the found block set
						matchingBlockSets.iterator().next().addAll( newSet );
					}
					else {
						// create new block set
						blockSetsToMerge.add( newSet );
					}	
				}
				
				LinkedList<Integer> dirtRows = new LinkedList<Integer>();
				
				// move all found unmatched rows/cols to their stored blocks
				for ( Pair<Integer, IndexBlock> p : colsToMerge ) 
					p.second.cols.add( p.first );
				
				for ( Pair<Integer, IndexBlock> p : rowsToMerge ) 
					p.second.rows.add( p.first );

				
				// merge all found block sets
				for ( HashSet<IndexBlock> blockSet : blockSetsToMerge ) {
					IndexBlock newBlock = new IndexBlock();
					
					for ( IndexBlock b : blockSet ) {
						newBlock.rows.addAll( b.rows );
						newBlock.cols.addAll( b.cols );
					}
					
					// sort the rows and cols to counter the use of HashSet
					newBlock.rows.sort();
					newBlock.cols.sort();
					
					matchingResult.removeAll( blockSet );
					matchingResult.add( newBlock );
				}
				
				// move the dirt rows/cols to the dirt
				for ( Integer row : rowMatching.dirtIndices )
					dirtRows.add( row );
				
				for ( Integer col : columnMatching.dirtIndices ) {
					blocksToRemove.clear();
					
					for ( IndexBlock b : matchingResult ) {
						List<Integer> rowsToRemove = new LinkedList<Integer>();
						
						for ( int row : b.rows.data ) {
							if ( matrix[row][col] == true )
								if ( !dirtRows.contains( row ) ) {
									rowsToRemove.add( row );
									dirtRows.add( row );
								}
						}
						
						b.rows.removeAll( rowsToRemove );
					}
					
					matchingResult.removeAll( blocksToRemove );
				}
				
				// magical step: execute the remaining steps with several values for step 3 to find the configuration with the least dirt (preferably 1)
				// assume that the dirt-generating function depending on the chosen threshold value is monotonically decreasing
				double threshold = 0, bestThreshold = 0;
				double stepSize = 2;
				int step = 0;
				
				List<Integer> tempDirt = new LinkedList<Integer>();
				ArrayList<Integer> lastDirtAboveZero = new ArrayList<Integer>();
				LinkedList<IndexBlock> mergeResult = new LinkedList<IndexBlock>(), lastMergeResultAboveZero = new LinkedList<IndexBlock>();
				
				step = maxSteps;
				threshold = 1;

				do {
					// clone the matching result and dirt rows to have a independent data set
					tempDirt.clear();
					mergeResult.clear();
					
					tempDirt.addAll( dirtRows );
					
					for ( IndexBlock block : matchingResult )
						mergeResult.add( block.clone() );
					
					// step 3: merge all blocks connected by off-diagonal bits where the column of the bit contains enough other bits to cover the second block at least to the given threshold;
					//         if this is the case, all blocks that have off-diagonal bits in the row of the initial bit are merged with the two initial blocks also
					// e.g. (threshold of 1):
					//		1				1				1
					//		x 2				1 1				1 1
					//		  2			->	  1			->	  1
					//		x   3			x   3			1 1 1
					//		    3				3			1   1
					//		x x	  4			1 1   1			x	  2
					//		x     4			1     1				  2
					// blocks 1 and 4 would be merged, since block 4 is completely covered; 
					// blocks 2 and 3 are not completely covered, but block 2 is connected to block 4 through row 6, thus will be merged also 
					List<List<IndexBlock>> blocksToMerge = matchConnectedBlocks( matrix, mergeResult, tempDirt, threshold );
					
					// merge all found block sets
					for ( List<IndexBlock> blockList : blocksToMerge ) {
						IndexBlock newBlock = new IndexBlock();
						
						for ( IndexBlock b : blockList ) {
							newBlock.rows.addAll( b.rows );
							newBlock.cols.addAll( b.cols );
						}
						
						mergeResult.removeAll( blockList );
						mergeResult.add( newBlock );
					}
										
					// step 4: move all rows with remaining off-diagonal bits into the dirt;
					// in the example of step 3, row 6 (3rd picture) would be moved to the dirt:
					//		1				1				1				x     x
					//		x 2				1 1				1 1				1
					//		  2			->	  1			->	  1			->	1 1
					//		x   3			x   3			1 1 1			  1
					//		    3				3			1   1			1 1 1
					//		x x	  4			1 1   1			x	  2			1   1
					//		x     4			1     1				  2				  2
					LinkedList<Integer> rowsToMove = new LinkedList<Integer>();
					blocksToRemove.clear();
	
					for ( IndexBlock b : mergeResult ) {
						rowsToMove.clear();
						
						for ( int row : b.rows.data )
							loop: for ( int col = 0; col < matrix[0].length; ++col )
								if ( matrix[row][col] == true ) {
									// check if the found column is part of the block
									for ( int blockCol : b.cols.data )
										if ( col == blockCol )
											continue loop;
									
									// when this point is reached, the column is not part of the block, thus the whole row has to be moved to the dirt
									rowsToMove.add( row );
									
									break;	// continue with the next row
								}
						
						tempDirt.addAll( rowsToMove );
						b.rows.removeAll( rowsToMove );
					}
	
					// step 5: move all rows from the dirt, that cover only one block, into this very block
					LinkedList<Integer> rowsToRemove = new LinkedList<Integer>();
					IndexBlock block;
					
					loop1: for ( Integer row : tempDirt ) {
						block = null;
						
						loop2: for ( IndexBlock b : mergeResult ) {
							for ( int blockCol : b.cols.data )
								// row covers this block
								if ( matrix[row][blockCol] == true )
									if ( block == null ) {
										block = b;
										continue loop2;
									}
									else // found already a block, so this row can not be moved
										continue loop1;
							}
						
						// found a block, so move the row to this block
						if ( block != null ) {
							rowsToRemove.add( row );
							block.rows.add( row );
							
							// add all columns which are touched by the given row but not contained already in the block
							LinkedList<Integer> colsToAdd = new LinkedList<Integer>();
							
							for ( int col = 0; col < matrix[0].length; ++col )
								if ( matrix[row][col] == true && !block.cols.contains( col ) )
									colsToAdd.add( col );
									
							block.cols.addAll( colsToAdd );
						}
					}
					
					tempDirt.removeAll( rowsToRemove );
	
					// remove all empty blocks
					for ( IndexBlock b : mergeResult ) {
						if ( b.rows.isEmpty() )
							blocksToRemove.add( b );
					}
	
					mergeResult.removeAll( blocksToRemove );
					
					// fast check: if a threshold of 0 is not able to produce a better result than the best found so far, 
					// (presumably) no other threshold for this configuration will be better
					if ( threshold == 0.0 && bestResult != null && tempDirt.size() > bestResult.dirt.rows.size() )
						break;
					
					// store the last result above zero, if necessary
					if ( tempDirt.size() != 0 && lastDirtAboveZero.hashCode() != tempDirt.hashCode() ) {
						lastDirtAboveZero.clear();
						lastDirtAboveZero.addAll( tempDirt );
						
						lastMergeResultAboveZero.clear();
						lastMergeResultAboveZero.addAll( mergeResult );
						
						bestThreshold = threshold;
					}
					
					// check the results and modify the threshold accordingly
					stepSize /= 2.0;
					
					if ( tempDirt.size() == 0 && stepSize < 0 )
						stepSize = -stepSize;
					else if ( tempDirt.size() > 1 && stepSize > 0 )
						stepSize = -stepSize;
					
					threshold += stepSize;	
				} while ( tempDirt.size() != 1 && ++step < maxSteps && threshold >= 0 && threshold <= 1 );
				
				// set the last dirt above zero value to the current dirt value if no dirt has been found at all (in any of the steps)
				if ( lastDirtAboveZero.size() == 0 && lastMergeResultAboveZero.size() == 0 ) {
					lastDirtAboveZero.addAll( tempDirt );
					lastMergeResultAboveZero.addAll( mergeResult );
					bestThreshold = threshold;
				}
				
				// store the result
				if ( bestResult == null ) {
					bestResult = new OptimizedIndexBlock();
					
					bestResult.dirt = new IndexBlock();
					bestResult.dirt.rows = new IndexBlockArrayList( bestResult.dirt, lastDirtAboveZero );

					bestResult.subblocks = lastMergeResultAboveZero;
					
					bestResult.rowThreshold = rowMatching.threshold;
					bestResult.columnThreshold = columnMatching.threshold;
					bestResult.coverageThreshold = bestThreshold;
				}
				else if ( lastDirtAboveZero.size() > 0 ) {
					if ( bestResult.dirt.rows.size() == 0 || lastDirtAboveZero.size() < bestResult.dirt.rows.size() ) {
						// results with less dirt are better; any result with dirt is better than a result without dirt
						bestResult.dirt.rows = new IndexBlockArrayList( bestResult.dirt, lastDirtAboveZero );
						bestResult.subblocks = lastMergeResultAboveZero;
						
						bestResult.rowThreshold = rowMatching.threshold;
						bestResult.columnThreshold = columnMatching.threshold;
						bestResult.coverageThreshold = bestThreshold;
					}
				}	
				
				// stop the search if a suitable result has been found 
				if ( bestResult.dirt.rows.size() == 1 )
					break optimizerLoop;	
			}
		}
		
		return bestResult;
	}

	/**
	 * @param matrix
	 * @param result
	 * @param unmatchedRows
	 * @param outBlockSetsToMerge
	 * @param outRowsToMerge
	 * @param outDirtRows
	 * @param matchingBlocks
	 * @param cover
	 * @param coverageThreshold
	 */
	private MatchingResult matchUnmatchedRows( 
			boolean[][] matrix, 
			LinkedList<IndexBlock> result, 
			List<Integer> unmatchedRows, 
			double coverageThreshold ) {
		MatchingResult matchingResult = new MatchingResult();
		
		matchingResult.threshold = coverageThreshold;
		
		LinkedList<IndexBlock> matchingBlocks = new LinkedList<IndexBlock>();

		for ( Integer unmatchedRow : unmatchedRows ) {
			matchingBlocks.clear();

			int colCount = 0;
			
			// find all matching blocks
			loop: for ( IndexBlock b : result ) {
				for ( int blockCol : b.cols.data )
					if ( matrix[unmatchedRow][blockCol] == true ) {
						matchingBlocks.add( b );
						colCount += b.rows.size();
						continue loop;
					}
			}
			
			if ( matchingBlocks.size() == 1 )	// just store the unmatched row if only one block was found
				matchingResult.indicesToMerge.add( new Pair<Integer, IndexBlock>( unmatchedRow, matchingBlocks.getFirst() ) );
			else if ( matchingBlocks.size() > 1 && colCount <= matrix[0].length * coverageThreshold ) {	// store the blocks to merge along with the row to merge
				// find all block sets to merge that already contain one of the newly found blocks
				LinkedList<HashSet<IndexBlock>> matchingBlockSets = new LinkedList<HashSet<IndexBlock>>();
				
				for ( HashSet<IndexBlock> set : matchingResult.blockSetsToMerge )
					for ( IndexBlock b : matchingBlocks )
						if ( set.contains( b ) ) {
							matchingBlockSets.add( set );
							break;
						}
				
				HashSet<IndexBlock> blockSet;
				
				if ( matchingBlockSets.size() > 1 ) {
					// merge block sets
					blockSet = new HashSet<IndexBlock>();
					
					for ( HashSet<IndexBlock> set : matchingBlockSets )
						blockSet.addAll( set );
					
					blockSet.addAll( matchingBlocks );
					
					matchingResult.blockSetsToMerge.removeAll( matchingBlockSets );
					matchingResult.blockSetsToMerge.add( blockSet );
				}
				else if ( matchingBlockSets.size() == 1 ) {
					// just add the found blocks to the found block set
					blockSet = matchingBlockSets.iterator().next();
					
					blockSet.addAll( matchingBlocks );
				}
				else {
					// create new block set
					blockSet = new HashSet<IndexBlock>();
					
					blockSet.addAll( matchingBlocks );
					matchingResult.blockSetsToMerge.add( blockSet );
				}
				
				// store the column to match
				matchingResult.indicesToMerge.add( new Pair<Integer, IndexBlock>( unmatchedRow, blockSet.iterator().next() ) );
			}
			else 	// store the row as dirt row
				matchingResult.dirtIndices.add( unmatchedRow );
		}
		
		return matchingResult;
	}
		
	/**
	 * @param matrix
	 * @param result
	 * @param unmatchedCols
	 * @param coverageThreshold
	 */
	private MatchingResult matchUnmatchedColumns( 
			boolean[][] matrix, 
			LinkedList<IndexBlock> result, 
			List<Integer> unmatchedCols, 
			double coverageThreshold ) {
		MatchingResult matchingResult = new MatchingResult();
		
		matchingResult.threshold = coverageThreshold;
		
		LinkedList<IndexBlock> matchingBlocks = new LinkedList<IndexBlock>();
		LinkedList<Integer> cover = new LinkedList<Integer>();
		
		for ( Integer unmatchedCol : unmatchedCols ) {
			matchingBlocks.clear();
			cover.clear();
			
			// get all rows covered by this column
			for ( int row = 0; row < matrix.length; ++row )
				if ( matrix[row][unmatchedCol] == true )
					cover.add( row );
			
			// find all matching blocks
			loop: for ( IndexBlock b : result ) {
				for ( int blockRow : b.rows.data )
					for ( Integer coveredRow : cover )
						if ( coveredRow.equals( blockRow ) ) {
							matchingBlocks.add( b );
							continue loop;
						}
			}
			
			int rowCount = 0;
			
			for ( IndexBlock b : matchingBlocks )
				rowCount += b.rows.size();
			
			if ( matchingBlocks.size() == 1 )	// just store the unmatched column if only one block was found
				matchingResult.indicesToMerge.add( new Pair<Integer, IndexBlock>( unmatchedCol, matchingBlocks.getFirst() ) );
			else if ( matchingBlocks.size() > 1 && rowCount <= matrix.length * coverageThreshold ) {	// store the blocks to merge along with the column to merge
				// find all block sets to merge that already contain one of the newly found blocks
				LinkedList<HashSet<IndexBlock>> matchingBlockSets = new LinkedList<HashSet<IndexBlock>>();
				
				for ( HashSet<IndexBlock> set : matchingResult.blockSetsToMerge )
					for ( IndexBlock b : matchingBlocks )
						if ( set.contains( b ) ) {
							matchingBlockSets.add( set );
							break;
						}
				
				HashSet<IndexBlock> blockSet;

				if ( matchingBlockSets.size() > 1 ) {
					// merge block sets
					blockSet = new HashSet<IndexBlock>();
					
					for ( HashSet<IndexBlock> set : matchingBlockSets )
						blockSet.addAll( set );
					
					blockSet.addAll( matchingBlocks );
					
					matchingResult.blockSetsToMerge.removeAll( matchingBlockSets );
					matchingResult.blockSetsToMerge.add( blockSet );
				}
				else if ( matchingBlockSets.size() == 1 ) {
					// just add the found blocks to the found block set
					blockSet = matchingBlockSets.iterator().next();
					
					blockSet.addAll( matchingBlocks );
				}
				else {
					// create new block set
					blockSet = new HashSet<IndexBlock>();
					
					blockSet.addAll( matchingBlocks );
					matchingResult.blockSetsToMerge.add( blockSet );
				}
				
				// store the column to match
				matchingResult.indicesToMerge.add( new Pair<Integer, IndexBlock>( unmatchedCol, blockSet.iterator().next() ) );
			}
			else 	// store the column as dirt column
				matchingResult.dirtIndices.add( unmatchedCol );
		}
		
		return matchingResult;
	}
	
	/**
	 * Selects the block from the given list that matches the given block.
	 * @param blocks
	 * @param s
	 * @return
	 */
	private static IndexBlock selectBlock( IndexBlock[] blocks, IndexBlock s ) {
		for ( IndexBlock block : blocks )
			if ( block.equalsIndexBlock( s ) )
				return block;
				
		return null;
	}
	
	/**
	 * Tries to match blocks that are connected by an off-diagonal bit at least to the given threshold.
	 * @param matrix
	 * @param result
	 * @param dirtRows
	 * @param coverageThreshold
	 */
	private LinkedList<List<IndexBlock>> matchConnectedBlocks( boolean[][] matrix, LinkedList<IndexBlock> result, List<Integer> dirtRows, double coverageThreshold ) {
		LinkedList<List<IndexBlock>> blocksToMerge = new LinkedList<List<IndexBlock>>();
		//IndexBlock block;
		
		// validate all block hashes to save time later on
		for ( IndexBlock b : result )
			b.validateHash();
		
		// find all connected blocks and store them for later merging
		int blockIndex = 0;		
		int resultSize = result.size();
		
		//for ( int blockIndex = 0; blockIndex < size; ++blockIndex ) {
		for ( IndexBlock block : result ) {
			IndexBlock pconnectedBlock = null;
			
			// check if the two blocks are connected
			loop: for ( int blockIndex2 = blockIndex + 1; blockIndex2 < resultSize && pconnectedBlock == null; ++blockIndex2 ) {
				loop2: for ( int block2Row : result.get( blockIndex2 ).rows.data ) {
					// filter out dirt bits, because they are not of interest
					for ( Integer dirtRow : dirtRows )
						if ( block2Row == dirtRow.intValue() )
							continue loop2;

					// search for an off-diagonal bit
					for ( int blockCol : block.cols.data ) {
						if ( matrix[block2Row][blockCol] == true ) {
							// found an off-diagonal bit
							pconnectedBlock = result.get( blockIndex2 );
							
							// check if enough rows of the connected block are covered by off-diagonal bits in the column of the current bit								
							int count = 0;
							
							for ( int prow : pconnectedBlock.rows.data )
								if ( matrix[prow][blockCol] == true )
									++count;

							// if the connected block is covered enough, store it for merging but don't merge yet to avoid order dependencies
							if ( count >= pconnectedBlock.rows.size() * coverageThreshold ) {
								mergeBlockSets( blocksToMerge, block, pconnectedBlock );
								
								// merge all blocks into the current set, which are touched by any bit in the current row
								for ( IndexBlock b : result )	// find the corresponding block
									if ( b != block && b != pconnectedBlock )
										for ( int bcol : b.cols.data )
											if ( matrix[block2Row][bcol] == true ) {
												mergeBlockSets( blocksToMerge, block, b );
												break;
											}
								
								continue loop;
							}
						}
					}
				}
			}
						
			++blockIndex;
		}
		
		return blocksToMerge;
	}
	
	/**
	 * Merges to blocks in the given merge set, merging also sets if they overlap
	 * @param blocksToMerge
	 * @param block1
	 * @param block2
	 */
	private void mergeBlockSets( List<List<IndexBlock>> blocksToMerge, IndexBlock block1, IndexBlock block2 ) {
		List<IndexBlock> list1 = null;
		List<IndexBlock> list2 = null;
		
		int hash1 = block1.hash;
		int hash2 = block2.hash;
		
		loop: for ( List<IndexBlock> mergeSet : blocksToMerge ) {
			for ( IndexBlock b : mergeSet )
				// the hash codes of the intern blocks are all validated at this point, so the intern hash code function can be used directly
				if ( b.hash == hash1 ) {
					list1 = mergeSet;
					break loop;
				}
		}
			
		loop: for ( List<IndexBlock> mergeSet : blocksToMerge ) {
			for ( IndexBlock b : mergeSet )
				// the hash codes of the intern blocks are all validated at this point, so the intern hash code function can be used directly
				if ( b.hash == hash2 ) {
					list2 = mergeSet;
					break loop;
				}
		}
		
		if ( list1 == null && list2 == null ) {			// blocks not found yet
			List<IndexBlock> newList = new LinkedList<IndexBlock>();
			
			newList.add( block1 );
			
			if ( block2 != block1 )
				newList.add( block2 );
			
			blocksToMerge.add( newList );
		}
		else if ( list1 != null && list2 == null ) {	// first block already found
			list1.add( block2 );
		}
		else if ( list1 == null && list2 != null ) {	// second block already found
			list2.add( block1 );
		}
		else if ( list1 != list2 ) {					// both blocks found in separate sets; merge them
			blocksToMerge.remove( list2 );			

			for ( IndexBlock b : list2 )				// should (in theory) be possible to handle it through addAll, but better be sure... 
				if ( !list1.contains( b ) )
					list1.add( b );
		}
	}

	private static final boolean BFS_WHITE = true;
	private static final boolean BFS_GRAY = false;
	
	private int[][]			flow, capacity, res_capacity;
	private int[]			parent, queue;
	private boolean[]		color;
	private int[]			min_capacity;
	private int				size, source, sink, first, last;
	
	/**
	 * Calculates the maximum matching of the bipartite graph represented
	 * by the given coverage matrix.
	 * @param coverageMatrix
	 * @return list of Integer-pairs forming the maximum matching
	 */
	private LinkedList<RowColPair> calcMaximumMatching( boolean[][] coverageMatrix ) {

		if ( coverageMatrix.length == 0 )
			return new LinkedList<RowColPair>();

		size = coverageMatrix.length + coverageMatrix[0].length + 2; // two additional nodes source s and sink t

		// set the capacity
		// remark: the columns of the coverage matrix are placed behind the rows in the capacity matrix
		capacity = new int[size][size];

		int numRows = coverageMatrix.length;

		for ( int row = 0; row < numRows; ++row )
			for ( int col = 0; col < coverageMatrix[0].length; ++col ) {
				capacity[row][numRows + col] = coverageMatrix[row][col] ? 1 : 0;
				capacity[numRows + col][row] = coverageMatrix[row][col] ? -1 : 0;
			}

		// add the capacity for the source and sink
		for ( int row = 0; row < coverageMatrix.length; ++row ) {
			capacity[size - 2][row] = 1;
			capacity[row][size - 2] = -1;
		}

		for ( int col = 0; col < coverageMatrix[0].length; ++col ) {
			capacity[numRows + col][size - 1] = 1;
			capacity[size - 1][numRows + col] = -1;
		}

		source = size - 2;
		sink = size - 1;
				
		maxFlow();
		
		// extract the flow between the rows and columns (source and sink are ignored), which corresponds to the maximum matching
		LinkedList<RowColPair> matching = new LinkedList<RowColPair>();
		
		for ( int row = 0; row < numRows; ++row )
			for ( int col = 0; col < coverageMatrix[0].length; ++col )
				if ( flow[row][numRows+col] == 1 )
					matching.add( new RowColPair( row, col ) );
				
		return matching;
	}

	private void maxFlow() // Edmonds-Karp algorithm with O(VE) complexity
	{
		flow = new int[size][size];
		res_capacity = new int[size][size];
		parent = new int[size];
		min_capacity = new int[size];
		color = new boolean[size];
		queue = new int[size];

		for ( int i = 0; i < size; i++ )
			for ( int j = 0; j < size; j++ )
				res_capacity[i][j] = capacity[i][j];

		while ( BFS( source ) ) {
			int v = sink, u;
			while ( v != source ) {
				u = parent[v];
				flow[u][v] += min_capacity[sink];
				flow[v][u] -= min_capacity[sink];
				res_capacity[u][v] -= min_capacity[sink];
				res_capacity[v][u] += min_capacity[sink];
				v = u;
			}
		}
	}

	private boolean BFS( int source ) // Breadth First Search in O(V<sup>2</sup>)
	{
		for ( int i = 0; i < size; i++ ) {
			color[i] = BFS_WHITE;
			min_capacity[i] = Integer.MAX_VALUE;
		}

		first = last = 0;
		queue[last++] = source;
		color[source] = BFS_GRAY;

		while ( first != last ) // While "queue" not empty..
		{
			int v = queue[first++];
			for ( int u = 0; u < size; u++ )
				if ( color[u] == BFS_WHITE && res_capacity[v][u] > 0 ) {
					min_capacity[u] = Math.min( min_capacity[v], res_capacity[v][u] );
					parent[u] = v;
					color[u] = BFS_GRAY;
					if ( u == sink )
						return true;
					queue[last++] = u;
				}
		}
		return false;
	}
}
